<!doctype html>
<html lang="en">
    <head>
        <link rel="stylesheet" type="text/css" href="static/core.css">
    </head>
    <body>
        <main>
            {{ rendered_markdown | safe }}
        </main>
        <footer>
            {{ footnote }}
        </footer>
    </body>
</html>
